#include "php.h"
#include "zend_compile.h"
#include "php_ini.h"
#include "php_globals.h"
#include "php_orbit.h"
#include "phorb.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct _SystemExceptionRec SystemExceptionRec;

struct _SystemExceptionRec {
	const char	*repoid;
	const char	*package;
	const char	*text;
};

#define SYS_EX(name, exp) \
	{ "IDL:omg.org/CORBA/" name ":1.0", "CORBA" CLASS_SEPARATOR name, exp }

static SystemExceptionRec system_exceptions[] = {
	SYS_EX("SystemException", "Unspecified system exception"),
	SYS_EX("UNKNOWN", "The unknown exception"),
	SYS_EX("BAD_PARAM", "An invalid parameter was passed"),
	SYS_EX("NO_MEMORY", "Dynamic memory allocation failure"),
	SYS_EX("IMP_LIMIT", "Violated implementation limit"),
	SYS_EX("COMM_FAILURE", "Communication failure"),
	SYS_EX("INV_OBJREF", "Invalid object reference"),
	SYS_EX("NO_PERMISSION", "No permission for attempted operation"),
	SYS_EX("INTERNAL", "ORB internal error"),
	SYS_EX("MARSHAL", "Error marshalling parameter or result"),
	SYS_EX("INITIALIZE", "ORB initialization failure"),
	SYS_EX("NO_IMPLEMENT", "Operation implementation unavailable"),
	SYS_EX("BAD_TYPECODE", "Bad typecode"),
	SYS_EX("BAD_OPERATION", "Invalid operation"),
	SYS_EX("NO_RESOURCES", "Insufficient resources for request"),
	SYS_EX("NO_RESPONSE", "Response to request not yet available"),
	SYS_EX("PERSIST_STORE", "Persistant storage failure"),
	SYS_EX("BAD_INV_ORDER", "Routine invocations out of order"),
	SYS_EX("TRANSIENT", "Transient failure - reissue request"),
	SYS_EX("FREE_MEM", "Cannot free memory"),
	SYS_EX("INV_IDENT", "Invalid identifier syntax"),
	SYS_EX("INV_FLAG", "Invalid flag was specified"),
	SYS_EX("INTF_REPOS", "Error accessing interface repository"),
	SYS_EX("BAD_CONTEXT", "Error processing context"),
	SYS_EX("OBJ_ADAPTER", "Failure detected by object adapter"),
	SYS_EX("DATA_CONVERSION", "Data conversion error"),
	SYS_EX("OBJECT_NOT_EXIST", "Non-existent object, delete reference"),
	SYS_EX("TRANSACTION_REQUIRED", "Transaction required"),
	SYS_EX("TRANSACTION_ROLLBACK", "Transaction rolled back"),
	SYS_EX("INVALID_TRANSACTION", "Invalid transaction"),
	SYS_EX("INV_POLICY", "Invalid policy"),
	SYS_EX("CODESET_INCOMPATIBLE", "Codeset incompatible")
};

#undef SYS_EX

void phorb_throw(zval *value)
{
	HashPosition pos;
	ulong index;
	OrbSLS_FETCH();

	if (OrbSG(exception)) {
		zval_ptr_dtor(&OrbSG(exception));
		zend_error(E_WARNING, "CORBA exception received while "
		                      "holding an unhandled one.");
	}

	OrbSG(exception) = value;
	OrbSG(exception_counter)++;
}

void phorb_catch(void)
{
	zval *eval;
	zval *function;
	zval *res = NULL;
	zval **argv[1];
	OrbSLS_FETCH();

	if (!OrbSG(exception))
		return;

	eval = OrbSG(exception);
	OrbSG(exception) = NULL;

	if (!OrbSG(exception_handler)) {
		zend_error(E_WARNING, "Unhandled CORBA exception received.");
		goto error;
	}

	MAKE_STD_ZVAL(function);
	ZVAL_STRING(function, OrbSG(exception_handler), 1);
	argv[0] = &eval;

	call_user_function_ex(CG(function_table), NULL, function,
	                      &res, 1, argv, 0, NULL);

	if (res)
		zval_ptr_dtor(&res);

	zval_dtor(function);
	FREE_ZVAL(function);

error:
	zval_ptr_dtor(&eval);
}

int phorb_reset_exception_counter(void)
{
	int old;
	OrbSLS_FETCH();

	old = OrbSG(exception_counter);
	OrbSG(exception_counter) = 0;

	return old;
}

const char *phorb_find_exception(phorbIDL *idl, const char *repoid)
{
	GSList *list;
	char *ret = NULL;
	OrbSLS_FETCH();

	if (idl)
		ret = g_hash_table_lookup(idl->exceptions, repoid);
	else {
		for(list = OrbSG(idls); list; list = list->next) {
			idl = (phorbIDL *)list->data;
			ret = g_hash_table_lookup(idl->exceptions, repoid);
			if (ret)
				break;
		}
	}

	return ret;
}

void phorb_shutdown_exceptions(void)
{
	OrbSLS_FETCH();

	if (OrbSG(exception_handler)) {
		g_free(OrbSG(exception_handler));
		OrbSG(exception_handler) = NULL;
	}
	if (OrbSG(exception)) {
		zval_ptr_dtor(&OrbSG(exception));
		OrbSG(exception) = NULL;
	}
}

void phorb_setup_exception(phorbIDL *idl, const char *repoid,
                           const char *pkg, const char *parent)
{
	OrbSLS_FETCH();

	if (phorb_find_exception(idl, repoid))
		return;

	g_hash_table_insert(idl->exceptions, g_strdup(repoid), g_strdup(pkg));
}

void phorb_setup_exceptions(phorbIDL *idl)
{
	int i;

	for(i = 1; i < G_N_ELEMENTS(system_exceptions); i++)
		phorb_setup_exception(idl, system_exceptions[i].repoid,
		                      system_exceptions[i].package,
		                      "CORBA" CLASS_SEPARATOR
		                      "SystemException");

	phorb_setup_exception(idl, "IDL:omg.org/CORBA/SystemException:1.0",
	                      "CORBA" CLASS_SEPARATOR "SystemException",
	                      "CORBA" CLASS_SEPARATOR "Exception");
	phorb_setup_exception(idl, "IDL:omg.org/CORBA/UserException:1.0",
	                      "CORBA" CLASS_SEPARATOR "UserException",
	                      "CORBA" CLASS_SEPARATOR "Exception");
}

void phorb_init_exceptions(void)
{
	OrbSLS_FETCH();

	OrbSG(exception_counter) = 0;
	OrbSG(exception_handler) = NULL;
	OrbSG(exception) = NULL;
}

void phorb_exception_set_user_function(const char *fname)
{
	OrbSLS_FETCH();

	if (OrbSG(exception_handler))
		g_free(OrbSG(exception_handler));
	OrbSG(exception_handler) = g_strdup(fname);
}

zval *phorb_system_except(const char *repoid, CORBA_unsigned_long minor,
                          CORBA_completion_status status)
{
	const char *pkg = NULL;
	const char *text = NULL;
	char *status_str;
	char *tmp = NULL;
	zval *res = NULL;
	int c;
	int i;

	if (strncmp(repoid, "IDL:CORBA", 9) == 0)
		repoid = tmp = g_strconcat("IDL:omg.org/", &repoid[4], NULL);

	for(i = 0; i < G_N_ELEMENTS(system_exceptions); i++) {
		if (!strcmp(repoid, system_exceptions[i].repoid)) {
			pkg = system_exceptions[i].package;
			text = system_exceptions[i].text;
			break;
		}
	}

	if (tmp)
		g_free(tmp);

	if (!pkg) {
		g_warning("Unable to find system exception "
		          "with repoid `%s'", repoid);
		pkg = system_exceptions[0].package;
		text = system_exceptions[0].text;
	}

	switch(status) {
	case CORBA_COMPLETED_YES:
		status_str = "COMPLETED_YES";
		break;
	case CORBA_COMPLETED_NO:
		status_str = "COMPLETED_NO";
		break;
	case CORBA_COMPLETED_MAYBE:
		status_str = "COMPLETED_MAYBE";
		break;
	}

	MAKE_STD_ZVAL(res);
	array_init(res);
	add_assoc_string(res, "_class", (char *)pkg, 1);
	add_assoc_string(res, "text", (char *)text, 1);
	add_assoc_string(res, "status", status_str, 1);
	convert_to_object(res);

	return res;
}

zval *phorb_user_except(const char *repoid, zval *value)
{
	char *pkg;

	pkg = (char *)phorb_find_exception(NULL, repoid);

	if (!pkg) {
		if (value) {
			zval_dtor(value);
			FREE_ZVAL(value);
		}

		g_warning("Unknown repoid '%s' in user exception", repoid);

		return phorb_system_except("IDL:omg.org/CORBA/UNKNOWN:1.0",
		                           0, CORBA_COMPLETED_MAYBE);
	}

	add_assoc_string(value, "_class", pkg, 1);
	convert_to_object(value);

	return value;
}
